//The headers
#include "SDL/SDL.h"
#include "SDL/SDL_image.h"
#include "SDL/SDL_rotozoom.h"
#include <string>

//Screen attributes
const int SCREEN_WIDTH = 800;
const int SCREEN_HEIGHT = 600;
const int SCREEN_BPP = 32;

//The frame rate
const int FRAMES_PER_SECOND = 20;

const int BRICK_WIDTH = 15;
const int BRICK_HEIGHT = 15;

const int CHAR_WIDTH = 25;
const int CHAR_HEIGHT = 26;
const int CHAR_WALK_SPEED = 15;
const int CHAR_ROLL_DIST = 150;
const int CHAR_SCALE = 4;

const int CHAR_LEFT = 1;
const int CHAR_RIGHT = 2;
const int CHAR_CROUCH_LEFT =3;

const int GRAVITY = 1;
const bool SMOOTH_IMAGE = 0;

const int MOVE_LEFT = 1;
const int MOVE_RIGHT = 2;
const int MOVE_LEFT_STOP = 3;
const int MOVE_RIGHT_STOP = 4;
const int MOVE_JUMP = 5;
const int MOVE_CROUCH = 6;
const int MOVE_CROUCH_STOP = 7;


//The surfaces
SDL_Surface *screen = NULL;
SDL_Surface *misc = NULL;
SDL_Surface *character = NULL;
SDL_Surface *character_left = NULL;

//The event structure
SDL_Event event;

//sprites clips
SDL_Rect *cCharRunRight[ 4 ];
SDL_Rect *cCharRunLeft[ 4 ];

SDL_Rect *cCharJumpRight[ 7 ];
SDL_Rect *cCharLongJumpRight[ 2 ];
SDL_Rect *cCharFallRight;
SDL_Rect *cCharJumpLeft[ 7 ];
SDL_Rect *cCharLongJumpLeft[ 2 ];
SDL_Rect *cCharFallLeft;

SDL_Rect *cCharStandLeft;
SDL_Rect *cCharStandRight;

SDL_Rect *cCharPushedLeft[2];
SDL_Rect *cCharPushedRight[2];

SDL_Rect *cCharCrouchLeft[2];
SDL_Rect *cCharCrouchRight[2];

SDL_Rect *cCharRollLeft[5];
SDL_Rect *cCharRollRight[5];

SDL_Rect *cCharRopeClimb[3];

SDL_Rect cChar[36];

class player
{
    private:		

	int xVel,yVel;
	int xVelCom;
	bool crouchCom;	
	
	//frame to show
	int frame;
	int frame_special;
	
	//walking left or right
	int status;
	bool running;
	bool falling;
	bool rolling;
	bool crouched;
	bool climbing;
	
	int yJump,xRoll;
    
    public:		
	//The collision box of the square
    SDL_Rect box;
	
    //Initializes the variables
    void init(int x, int y);
    
    //Takes key presses and adjusts the square's velocity
    void Movement_Input(int movement);
    
	void Movement();
	
    //Shows the square on the screen
    void show();
};
player myPlayer;

class floor
{
    private:
    //The collision box of the square
    SDL_Rect box;
        
    public:
    //Initializes the variables
    void init();    
       
    //Shows the square on the screen
    void show();
};

//The timer
class Timer
{
    private:
    //The clock time when the timer started
    int startTicks;
    
    //The ticks stored when the timer was paused
    int pausedTicks;
    
    //The timer status
    bool paused;
    bool started;
    
    public:
    //Initializes variables
    Timer();
    
    //The various clock actions
    void start();
    void stop();
    void pause();
    void unpause();
    
    //Get the number of ticks since the timer started
    //or gets the number of ticks when the timer was paused
    int get_ticks();
    
    //Checks the status of the timer
    bool is_started();
    bool is_paused();    
};

void set_clips()
{
	int i;
	
    //Clip the sprites
	
	for(i=0;i<36;i++)
	{
    	cChar[i].x = CHAR_WIDTH*CHAR_SCALE *(i%12);
    	cChar[i].y = CHAR_HEIGHT*CHAR_SCALE *(i/12);
    	cChar[i].w = CHAR_WIDTH*CHAR_SCALE;
    	cChar[i].h = CHAR_HEIGHT*CHAR_SCALE;
    }
	
	cCharStandRight = &cChar[0];
	cCharStandLeft = &cChar[11];
	
	
	for(i=0;i<4;i++)
	{
    	cCharRunRight[i ] = &cChar[1+i];
		cCharRunLeft[ i ] = &cChar[10-i];
    }
	
	for(i=0;i<3;i++)
	{
    	cCharRopeClimb[i] = &cChar[24];
    }
	
	for(i=0;i<7;i++)
	{
		cCharJumpRight[ i ]= &cChar[12+i];
		cCharJumpLeft[ i ]= &cChar[23-i];
    }
	
	for(i=0;i<2;i++)
	{
		cCharLongJumpRight[ i ]= &cChar[19+i];
		cCharLongJumpLeft[ i ]= &cChar[16-i];
		
		cCharPushedRight[i]= &cChar[22+i];
		cCharPushedLeft[i]= &cChar[13-i];
		
		cCharCrouchRight[i]=&cChar[5+i];
		cCharCrouchLeft[i]=&cChar[6-i];
    }

	cCharFallRight = &cChar[21];
	cCharFallLeft = &cChar[14];
	
	for(i=0;i<5;i++)
	{
		cCharRollRight[i]=&cChar[7+i];
		cCharRollLeft[i]=&cChar[5-i];
	}
	
}

SDL_Surface *load_image( std::string filename, bool revert ) 
{
    //The image that's loaded
    SDL_Surface* loadedImage = NULL;
    
	
    //The optimized surface that will be used
    SDL_Surface* optimizedImage = NULL;
	
	//Zoomed one
	SDL_Surface* zoomedImage = NULL;
    
    //Load the image
    loadedImage = IMG_Load( filename.c_str() );
    
    //If the image loaded
    if( loadedImage != NULL )
    {
        //Create an optimized surface
        optimizedImage = SDL_DisplayFormat( loadedImage );
        
        //Free the old surface
        SDL_FreeSurface( loadedImage );
        
        //If the surface was optimized
        if( optimizedImage != NULL )
        {
			if(revert)
				zoomedImage = zoomSurface (optimizedImage,-CHAR_SCALE, CHAR_SCALE, SMOOTH_IMAGE);
            else
				zoomedImage = zoomSurface (optimizedImage,CHAR_SCALE, CHAR_SCALE, SMOOTH_IMAGE);
			//Color key surface
            SDL_SetColorKey( zoomedImage, SDL_RLEACCEL | SDL_SRCCOLORKEY, SDL_MapRGB( optimizedImage->format, 0xFF, 0xFF, 0xFF ) );
        }
    }
	


	
	
    //Return the optimized surface
    return zoomedImage;
}

void apply_surface( int x, int y, SDL_Surface* source, SDL_Surface* destination, SDL_Rect* clip = NULL )
{
    //Holds offsets
    SDL_Rect offset;
    
    //Get offsets
    offset.x = x;
    offset.y = y;
    
    //Blit
    SDL_BlitSurface( source, clip, destination, &offset );
}

bool check_collision( SDL_Rect &A, SDL_Rect &B )
{
    //The sides of the rectangles
    int leftA, leftB;
    int rightA, rightB;
    int topA, topB;
    int bottomA, bottomB;

    //Calculate the sides of rect A
    leftA = A.x;
    rightA = A.x + A.w;
    topA = A.y;
    bottomA = A.y + A.h;
        
    //Calculate the sides of rect B
    leftB = B.x;
    rightB = B.x + B.w;
    topB = B.y;
    bottomB = B.y + B.h;
            
    //If any of the sides from A are outside of B
    if( bottomA <= topB )
    {
        return false;
    }
    
    if( topA >= bottomB )
    {
        return false;
    }
    
    if( rightA <= leftB )
    {
        return false;
    }
    
    if( leftA >= rightB )
    {
        return false;
    }
    
    //If none of the sides from A are outside B
    return true;
}

bool init()
{
    //Initialize all SDL subsystems
    if( SDL_Init( SDL_INIT_EVERYTHING ) == -1 )
    {
        return false;    
    }
    
    //Set up the screen
    screen = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, SDL_SWSURFACE );
    
    //If there was in error in setting up the screen
    if( screen == NULL )
    {
        return false;    
    }
    
    //Set the window caption
    SDL_WM_SetCaption( "Marshmallow", NULL );
    
    //If everything initialized fine
    return true;
}

bool load_files()
{
    //Load the square image
	
	misc = load_image( "misc.bmp" ,0);
	character = load_image( "char.pcx" ,0);
	character_left = load_image( "char.pcx" , 1);
    
    //If there was a problem in loading the square
    if( character_left == NULL )
    {
        return false;    
    }
	if( character == NULL )
    {
        return false;    
    }
    
    //If everything loaded fine
    return true;
}

void clean_up()
{
    //Free the surface
    SDL_FreeSurface( misc );
	SDL_FreeSurface( character );
	SDL_FreeSurface( character_left );
    
    //Quit SDL
    SDL_Quit();
}


void player::init(int x,int y)
{
	box.x= x;
	box.y= y;
	
	box.w= CHAR_WIDTH*CHAR_SCALE;
	box.h= CHAR_HEIGHT*CHAR_SCALE;
	
	xVel = 0;
	xVelCom = 0;
	yVel = 0;
	
	frame = 0;
	frame_special = 0;
	status = CHAR_RIGHT;
	
	running = 0;
	falling = 0;
	rolling = 0;
	climbing = 0;
	
	yJump = 0;
	xRoll = 0;
}

void player::show()
{
//	printf("crouched=%d rolling=%d running=%d xVel=%d falling=%d crouchCom=%d xVelCom=%d\n",
//							crouched,rolling,running,xVel,falling,crouchCom,xVelCom);
	
	if(crouched)
		xVel=0;
	
	if((!crouched) && (!rolling) && (!falling))
		xVel = xVelCom;
	
	if((!falling) && (!rolling) && (!running))
		crouched=crouchCom;
	
	if ((!falling) && (!rolling) && (xVel!=0) && crouchCom)
	{		
		rolling=1;
		xRoll=box.x;
	}

		
	//handle rolling
	if( (rolling) && (((box.x-xRoll)>=CHAR_ROLL_DIST) || ((box.x-xRoll)<(-CHAR_ROLL_DIST))))
	{	
		rolling=0;
		xVel=0;
		crouched=crouchCom;
	}	
	
    //Move the square left or right
    box.x += xVel;
       
    //If the square went too far to the left or right or has collided with the wall
    if( ( box.x < 0 ) || ( box.x + CHAR_WIDTH*CHAR_SCALE > SCREEN_WIDTH ))
    {
        //Move back
        box.x -= xVel;
		rolling=0;
		crouched=crouchCom;
    }
	
	//If Foo is moving left 
	if( xVel < 0 ) 
	{ 
		//Set the animation to left 
		status = CHAR_LEFT;
		running = 1;		
		//Move to the next frame in the animation 
		frame++; 
	}
	
	//If Foo is moving right
	else if( xVel > 0 ) 
	{ 
		//Set the animation to right
		status = CHAR_RIGHT;
		running = 1;			
		//Move to the next frame in the animation 
		frame++; 
		
	} //If Foo is standing
	else 
	{ 
		//Set standing 
		running = 0;
		//Restart the animation 		
		frame = 0; 
	}
	
	//Loop the animation
	if( frame >= 8 ) 
	{ 
		frame = 0;
	}
    
	
	//Move the square up or down
	if (falling)
	{
    	box.y += yVel;
    	yVel += GRAVITY;
	}
	
	if(yVel != 0)
	{
		falling = 1;
	}
	
    //If the square went too far up or down or has collided with the wall
    if(  box.y + CHAR_HEIGHT*CHAR_SCALE > SCREEN_HEIGHT )
    {
        box.y = SCREEN_HEIGHT-CHAR_HEIGHT*CHAR_SCALE;
		falling = 0;
		if(crouchCom)
			rolling=1;
		yVel = 0;
    }   
    
	if((!crouched) && (!falling) && (!rolling))
		frame_special=0;
	
	switch (status)
	{
		case CHAR_RIGHT:
			if(falling)
			{
				if(xVel == 0 && box.y < yJump)
					apply_surface( box.x, box.y, character, screen, cCharJumpRight[(frame_special++/3) %7]);	
				else if(box.y < yJump && yVel <= 0)
					apply_surface( box.x, box.y, character, screen, cCharLongJumpRight[0]);				
				else if(box.y < yJump && yVel > 0)
					apply_surface( box.x, box.y, character, screen, cCharLongJumpRight[1]);
				else 
					apply_surface( box.x, box.y, character, screen, cCharFallRight);
			}
			else if(crouched)
				apply_surface( box.x, box.y, character, screen, cCharCrouchRight[ (frame_special==0)?frame_special++:1 ]);			
			else if(rolling)
				apply_surface( box.x, box.y, character, screen, cCharRollRight[ frame_special++/2%5 ]);
			else if(running)
				apply_surface( box.x, box.y, character, screen, cCharRunRight[ frame/2 ]);
			else
				apply_surface( box.x, box.y, character, screen, cCharStandRight);
			break;
			
		case CHAR_LEFT:
			if(falling)
			{
				if(xVel == 0 && box.y < yJump)
					apply_surface( box.x, box.y, character_left, screen, cCharJumpLeft[(frame_special++/3) %7]);	
				else if(box.y < yJump && yVel <= 0)
					apply_surface( box.x, box.y, character_left, screen, cCharLongJumpLeft[0]);				
				else if(box.y < yJump && yVel > 0)
					apply_surface( box.x, box.y, character_left, screen, cCharLongJumpLeft[1]);
				else 
					apply_surface( box.x, box.y, character_left, screen, cCharFallLeft);
			}			
			else if(crouched)
				apply_surface( box.x, box.y, character_left, screen, cCharCrouchLeft[ (frame_special==0)?frame_special++:1 ]);
			else if(rolling)
				apply_surface( box.x, box.y, character_left, screen, cCharRollLeft[ frame_special++/2%5 ]);
			else if(running)
				apply_surface( box.x, box.y, character_left, screen, cCharRunLeft[ frame/2 ]);
			else
				apply_surface( box.x,box.y, character_left,screen, cCharStandLeft);
			break;
	}
	
}


void player::Movement_Input(int movement)
{
	switch (movement)
	{
		case MOVE_LEFT:
			xVelCom -= CHAR_WALK_SPEED;
			break;
				
		case MOVE_RIGHT:
			xVelCom += CHAR_WALK_SPEED;
			break;
			
		case MOVE_LEFT_STOP:
			xVelCom += CHAR_WALK_SPEED;
			break;
			
		case MOVE_RIGHT_STOP:			
			xVelCom -= CHAR_WALK_SPEED;
			break;
			
		case MOVE_JUMP:
			if(!falling && !rolling && !crouched)
			{
				yVel = -10;
				yJump = box.y;
			}
			break;
			
		case MOVE_CROUCH:
			crouchCom=1;
			break;
		
		case MOVE_CROUCH_STOP:
			crouchCom=0;
			break;
	}
}

Timer::Timer()
{
    //Initialize the variables
    startTicks = 0;
    pausedTicks = 0;
    paused = false;
    started = false;    
}

void Timer::start()
{
    //Start the timer
    started = true;
    
    //Unpause the timer
    paused = false;
    
    //Get the current clock time
    startTicks = SDL_GetTicks();    
}

void Timer::stop()
{
    //Stop the timer
    started = false;
    
    //Unpause the timer
    paused = false;    
}

void Timer::pause()
{
    //If the timer is running and isn't already paused
    if( ( started == true ) && ( paused == false ) )
    {
        //Pause the timer
        paused = true;
    
        //Calculate the paused ticks
        pausedTicks = SDL_GetTicks() - startTicks;
    }
}

void Timer::unpause()
{
    //If the timer is paused
    if( paused == true )
    {
        //Unpause the timer
        paused = false;
    
        //Reset the starting ticks
        startTicks = SDL_GetTicks() - pausedTicks;
        
        //Reset the paused ticks
        pausedTicks = 0;
    }
}

int Timer::get_ticks()
{
    //If the timer is running
    if( started == true )
    {
        //If the timer is paused
        if( paused == true )
        {
            //Return the number of ticks when the the timer was paused
            return pausedTicks;
        }
        else
        {
            //Return the current time minus the start time
            return SDL_GetTicks() - startTicks;
        }    
    }
    
    //If the timer isn't running
    return 0;    
}

bool Timer::is_started()
{
    return started;    
}

bool Timer::is_paused()
{
    return paused;    
}
    
void handle_input()
{
	if( event.type == SDL_KEYDOWN )
    {
        //Adjust the velocity
        switch( event.key.keysym.sym )
        {
            case SDLK_LEFT: myPlayer.Movement_Input(MOVE_LEFT); break;
            case SDLK_RIGHT: myPlayer.Movement_Input(MOVE_RIGHT); break;				
           	case SDLK_UP: myPlayer.Movement_Input(MOVE_JUMP); break;
			case SDLK_DOWN: myPlayer.Movement_Input(MOVE_CROUCH); break;				
        }
    }
    //If a key was released
    else if( event.type == SDL_KEYUP )
    {
        //Adjust the velocity
        switch( event.key.keysym.sym )
        {
            case SDLK_LEFT: myPlayer.Movement_Input(MOVE_LEFT_STOP); break;
            case SDLK_RIGHT: myPlayer.Movement_Input(MOVE_RIGHT_STOP); break;				
			case SDLK_DOWN: myPlayer.Movement_Input(MOVE_CROUCH_STOP); break;				
        }        
    }
}
int main( int argc, char* args[] )
{
    //Quit flag
    bool quit = false;
    
	myPlayer.init(0,0);
	
	//The frame rate regulator
    Timer fps;
    
    //Initialize
    if( init() == false )
    {
        return 1;
    }

    //Load the files
    if( load_files() == false )
    {
        return 1;
    }
    
	//set clips
    set_clips();
	
    //While the user hasn't quit
    while( quit == false )
    {
        //Start the frame timer
        fps.start();
        
        //While there's events to handle
        while ( SDL_PollEvent( &event ) )
        {
            //Handle events for the dot
            handle_input();
            
            //If the user has Xed out the window
            if( event.type == SDL_QUIT )
            {
                //Quit the program
                quit = true;
            }
        }
        
        //Fill the screen white
        SDL_FillRect( screen, &screen->clip_rect, SDL_MapRGB( screen->format, 0x0, 0x0, 0x0 ) );
		
		//Show the bats
		myPlayer.show();
        //Update the screen
        if( SDL_Flip( screen ) == -1 )
        {
            return 1;    
        }
        
        //Cap the frame rate
        while( fps.get_ticks() < 1000 / FRAMES_PER_SECOND )
        {
            //wait    
        }
    }
        
    //Clean up
    clean_up();
    
    return 0;    
}
